########################################################################
# File name: service.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio

import aioxmpp.disco
import aioxmpp.service
import aioxmpp.structs

from aioxmpp.utils import namespaces

from . import xso as ping_xso


class PingService(aioxmpp.service.Service):
    """
    Service implementing XMPP Ping (:xep:`199`).

    This service implements the response to XMPP Pings and provides a method
    to send pings.

    .. automethod:: ping
    """

    ORDER_AFTER = [
        aioxmpp.disco.DiscoServer,
    ]

    _ping_feature = aioxmpp.disco.register_feature(
        namespaces.xep0199_ping
    )

    @aioxmpp.service.iq_handler(aioxmpp.structs.IQType.GET, ping_xso.Ping)
    async def handle_ping(self, request):
        return ping_xso.Ping()

    async def ping(self, peer):
        """
        Wrapper around :func:`aioxmpp.ping.ping`.

        **When to use this wrapper vs. the global function:** Using this method
        has the side effect that the application will start to respond to
        :xep:`199` pings. While this is not a security issue per se (as
        responding to :xep:`199` pings only changes the format of the reply,
        not the fact that a reply is sent from the client), it may not be
        desirable under all circumstances.

        So especially when developing a Service which does not *reqiure* that
        the application replies to pings (for example, when implementing a
        stream or group chat aliveness check), it is preferrable to use the
        global function.

        When implementing an application where it is desirable to reply to
        pings, using this wrapper is fine.

        In general, aioxmpp services should avoid depending on this service.

        (The decision essentially boils down to "summon this service or not?",
        and it is not a decision aioxmpp should make for the application unless
        necessary for compliance.)

        .. versionchanged:: 0.11

            Converted to a shim wrapper.
        """
        return await ping(self.client, peer)


async def ping(client, peer):
    """
    Ping a peer.

    :param peer: The peer to ping.
    :type peer: :class:`aioxmpp.JID`
    :raises aioxmpp.errors.XMPPError: as received

    Send a :xep:`199` ping IQ to `peer` and wait for the reply.

    This is a low-level version of :meth:`aioxmpp.PingService.ping`.

    **When to use this function vs. the service method:** See
    :meth:`aioxmpp.PingService.ping`.

    .. note::

        If the peer does not support :xep:`199`, they will respond with
        a ``cancel`` ``service-unavailable`` error. However, some
        implementations return a ``cancel`` ``feature-not-implemented``
        error instead. Callers should be prepared for the
        :class:`aioxmpp.XMPPCancelError` exceptions in those cases.

    .. versionchanged:: 0.11

        Extracted this helper from :class:`aioxmpp.PingService`.
    """

    iq = aioxmpp.IQ(
        to=peer,
        type_=aioxmpp.IQType.GET,
        payload=ping_xso.Ping()
    )

    await client.send(iq)



########################################################################
# File name: pubsub_watch.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import itertools

import aioxmpp
import aioxmpp.pubsub
import aioxmpp.xml

from framework import Example, exec_example


class PubSubWatch(Example):
    def prepare_argparse(self):
        super().prepare_argparse()

        # this gives a nicer name in argparse errors
        def jid(s):
            return aioxmpp.JID.fromstr(s)

        self.argparse.add_argument(
            "target_entity",
            type=jid
        )

        self.argparse.add_argument(
            "target_node",
            default=None,
            nargs="?",
        )

    async def run_example(self):
        self.stop_event = self.make_sigint_event()
        await super().run_example()

    def _on_item_published(self, jid, node, item, *, message=None):
        print("PUBLISHED: {}".format(item.id_))

    def _on_item_retracted(self, jid, node, id_, *, message=None):
        print("RETRACTED: {}".format(id_))

    async def run_simple_example(self):
        pubsub = self.client.summon(aioxmpp.PubSubClient)
        pubsub.on_item_published.connect(self._on_item_published)
        pubsub.on_item_retracted.connect(self._on_item_retracted)
        subid = (await pubsub.subscribe(
            self.args.target_entity,
            node=self.args.target_node,
        )).payload.subid
        print("SUBSCRIBED: subid={!r}".format(subid))
        try:
            await self.stop_event.wait()
        finally:
            await pubsub.unsubscribe(
                self.args.target_entity,
                node=self.args.target_node,
                subid=subid,
            )


if __name__ == "__main__":
    exec_example(PubSubWatch())



########################################################################
# File name: __init__.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.misc` -- Miscellaneous XSOs
##########################################

This subpackage bundles XSO definitions for several XEPs. They do not get their
own subpackage because they often only define one or two XSOs without any logic
involved. The XSOs are often intended for re-use by other protocols.


Out of Band Data (:xep:`66`)
============================

.. autoclass:: OOBExtension

.. attribute:: aioxmpp.Message.xep0066_oob


Delayed Delivery (:xep:`203`)
=============================

.. autoclass:: Delay()

.. attribute:: aioxmpp.Message.xep0203_delay

   A :class:`Delay` instance which indicates that the message has been
   delivered with delay.


Stanza Forwarding (:xep:`297`)
==============================

.. autoclass:: Forwarded()


Last Message Correction (:xep:`308`)
====================================

.. autoclass:: Replace()

.. attribute:: aioxmpp.Message.xep308_replace

    A :class:`Replace` instance which indicates that the message is supposed
    to replcae another message.


Chat Markers (:xep:`333`)
=========================

.. autoclass:: ReceivedMarker

.. autoclass:: DisplayedMarker

.. autoclass:: AcknowledgedMarker

.. attribute:: aioxmpp.Message.xep0333_marker


JSON Containers (:xep:`335`)
============================

:xep:`335` defines a standard way to transport JSON data in XMPP. The
:class:`JSONContainer` is an XSO class which represents the ``<json/>`` element
specified in :xep:`335`.

:mod:`aioxmpp` also provides an :class:`~aioxmpp.xso.AbstractElementType`
called :class:`JSONContainerType` which can be used to extract JSON data from
an element using the :class:`JSONContainer` format.

.. autoclass:: JSONContainer

.. autoclass:: JSONContainerType


Pre-Authenticated Roster Subcription (:xep:`379`)
=================================================

.. autoclass:: Preauth

.. attribute:: aioxmpp.Presence.xep0379_preauth

   The pre-auth element associate with a subscription request.


Current Jabber OpenPGP Usage (:xep:`27`)
========================================

.. autoclass:: OpenPGPEncrypted

.. autoclass:: OpenPGPSigned

.. attribute:: aioxmpp.Message.xep0027_encrypted

    Instance of :class:`OpenPGPEncrypted`, if present.

    .. note::

        :xep:`27` does not specify the signing of messages.

.. attribute:: aioxmpp.Presence.xep0027_signed

    Instance of :class:`OpenPGPSigned`, if present.

"""

from .delay import Delay  # NOQA: F401
from .lmc import Replace  # NOQA: F401
from .forwarding import Forwarded  # NOQA: F401
from .oob import OOBExtension  # NOQA: F401
from .markers import (  # NOQA: F401
    ReceivedMarker,
    DisplayedMarker,
    AcknowledgedMarker,
)
from .json import JSONContainer, JSONContainerType  # NOQA: F401
from .pars import Preauth  # NOQA: F401
from .openpgp_legacy import (
    OpenPGPEncrypted,
    OpenPGPSigned,
)



########################################################################
# File name: test_xso.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import unittest

import aioxmpp
import aioxmpp.xso as xso
import aioxmpp.ibb.xso as ibb_xso

from aioxmpp.utils import namespaces


class TestNamespace(unittest.TestCase):
    def test_namespace(self):
        self.assertEqual(
            namespaces.xep0047,
            "http://jabber.org/protocol/ibb"
        )


class TestOpen(unittest.TestCase):

    def test_tag(self):
        self.assertEqual(
            ibb_xso.Open.TAG,
            (namespaces.xep0047, "open")
        )


class TestClose(unittest.TestCase):

    def test_tag(self):
        self.assertEqual(
            ibb_xso.Close.TAG,
            (namespaces.xep0047, "close")
        )


class TestData(unittest.TestCase):
    def test_tag(self):
        self.assertEqual(
            ibb_xso.Data.TAG,
            (namespaces.xep0047, "data")
        )

    def test_init(self):
        data = ibb_xso.Data("sessionid", 10, b"adsf\x02")
        self.assertEqual(data.sid, "sessionid")
        self.assertEqual(data.seq, 10)
        self.assertEqual(data.content, b"adsf\x02")


class TestMonkeyPatch(unittest.TestCase):

    def test_Message_attribute(self):
        self.assertIsInstance(
            aioxmpp.Message.xep0047_data,
            xso.Child,
        )
        self.assertSetEqual(
            aioxmpp.Message.xep0047_data._classes,
            {
                ibb_xso.Data,
            }
        )



########################################################################
# File name: test_e2e.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import unittest

import aioxmpp
import aioxmpp.ibb

from aioxmpp.testutils import get_timeout

from aioxmpp.e2etest import (
    blocking_timed,
    TestCase,
)


class TestProtocol(asyncio.Protocol):

    def __init__(self):
        self.data = b""
        self._transport = None
        self.connection_lost_fut = asyncio.Future()

    def connection_made(self, transport):
        self._transport = transport

    def connection_lost(self, e):
        self.connection_lost_fut.set_result(e)

    def pause_writing(self):
        pass

    def resume_writing(self):
        pass

    def data_received(self, data):
        self.data += data


class TestIBB(TestCase):
    async def make_client(self, run_before=None):
        return await self.provisioner.get_connected_client(
            services=[
                aioxmpp.ibb.IBBService,
            ],
            prepare=run_before,
        )

    @blocking_timed
    async def test_ibb(self):
        client1 = await self.make_client()
        client2 = await self.make_client()

        s1 = client1.summon(aioxmpp.ibb.IBBService)
        s2 = client2.summon(aioxmpp.ibb.IBBService)

        # set-up the session
        handle2_fut = s2.expect_session(
            TestProtocol, client1.local_jid, "fnord")
        transport1, proto1 = await s1.open_session(
            TestProtocol, client2.local_jid, sid="fnord")
        transport2, proto2 = await handle2_fut

        # transfer data
        transport2.write(b"this")
        transport2.write(b"is")
        transport2.write(b"data")
        transport2.close()

        # assert that both protocols get notified (otherwise time out)
        e1 = await proto1.connection_lost_fut
        self.assertIsNone(e1)
        e2 = await proto2.connection_lost_fut
        self.assertIsNone(e2)

        self.assertEqual(proto1.data, b"thisisdata")

    @blocking_timed
    async def test_ibb_message(self):
        client1 = await self.make_client()
        client2 = await self.make_client()

        s1 = client1.summon(aioxmpp.ibb.IBBService)
        s2 = client2.summon(aioxmpp.ibb.IBBService)

        # set-up the session
        handle2_fut = s2.expect_session(
            TestProtocol, client1.local_jid, "fnord")
        transport1, proto1 = await s1.open_session(
            TestProtocol, client2.local_jid, sid="fnord",
            stanza_type=aioxmpp.ibb.IBBStanzaType.MESSAGE
        )
        transport2, proto2 = await handle2_fut

        # transfer data
        transport2.write(b"this")
        transport2.write(b"is")
        transport2.write(b"data")
        transport2.close()

        # assert that both protocols get notified (otherwise time out)
        e1 = await proto1.connection_lost_fut
        self.assertIsNone(e1)
        e2 = await proto2.connection_lost_fut
        self.assertIsNone(e2)

        self.assertEqual(proto1.data, b"thisisdata")

    @blocking_timed
    async def test_ibb_client_disconnects(self):
        client1 = await self.make_client()
        client2 = await self.make_client()

        s1 = client1.summon(aioxmpp.ibb.IBBService)
        s2 = client2.summon(aioxmpp.ibb.IBBService)

        # set-up the session
        handle2_fut = s2.expect_session(
            TestProtocol, client1.local_jid, "fnord")
        transport1, proto1 = await s1.open_session(
            TestProtocol, client2.local_jid, sid="fnord")
        transport2, proto2 = await handle2_fut

        # transfer data
        transport2.write(b"this")
        transport1.abort()
        transport2.write(b"is")

        # assert that both protocols get notified (otherwise time out)
        e1 = await proto1.connection_lost_fut
        self.assertIsNone(e1)
        e2 = await proto2.connection_lost_fut
        self.assertIsInstance(e2, aioxmpp.errors.XMPPCancelError)



#!/usr/bin/env python3
########################################################################
# File name: pushbot.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import json
import logging
import pathlib
import socket
import signal

import toml

import aioxmpp


class MessageProtocol(asyncio.DatagramProtocol):
    def __init__(self, queue):
        super().__init__()
        self.logger = logging.getLogger(type(self).__name__)
        self.queue = queue

    def datagram_received(self, data, addr):
        try:
            parsed = json.loads(data.decode("utf-8"))
        except Exception:
            self.logger.error("failed to parse client message %r",
                              data,
                              exc_info=True)
            return

        try:
            self.queue.put_nowait(parsed)
        except asyncio.QueueFull:
            self.logger.error("input queue full! dropped message %r",
                              parsed)


async def process_item(item, rooms, logger):
    target_rooms = rooms.keys()

    tokens = []

    for room_address in target_rooms:
        try:
            room_info = rooms[room_address]
        except KeyError:
            continue

        body_parts = []

        if room_info["head_format"]:
            body_parts.append(
                room_info["head_format"].format(
                    nitems=len(item["items"]),
                    root_item=item,
                )
            )

        for sub_item in item["items"]:
            required_fields = room_info["required_fields"]
            if required_fields:
                item_fields = set(sub_item.keys()) & required_fields
                if len(item_fields) < len(required_fields):
                    continue

            format_ = room_info["format"]
            if format_:
                body = format_.format(**sub_item)
            else:
                body = repr(item)

            body_parts.append(body)

        msg = aioxmpp.Message(
            type_=aioxmpp.MessageType.GROUPCHAT,
        )
        msg.body[None] = "\n".join(body_parts)

        tokens.append(asyncio.ensure_future(
            room_info["room"].send_message(msg)
        ))

    if not tokens:
        logger.warning("item %r generated no message!", item)
        return

    await asyncio.wait(tokens, return_when=asyncio.ALL_COMPLETED)


async def process_queue(queue, rooms):
    logger = logging.getLogger("processor")

    while True:
        item = await queue.get()
        try:
            await process_item(item, rooms, logger)
        except Exception:
            logger.error("failed to process item!", exc_info=True)
            continue


async def amain(loop, xmpp_cfg, unix_cfg, mucs):
    message_queue = asyncio.Queue(maxsize=16)
    message_handler = MessageProtocol(message_queue)

    sigint_received = asyncio.Event()
    sigint_future = asyncio.ensure_future(sigint_received.wait())

    loop.add_signal_handler(signal.SIGINT, sigint_received.set)
    loop.add_signal_handler(signal.SIGTERM, sigint_received.set)

    socket_path, = unix_cfg

    sock = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM, 0)
    sock.bind(str(socket_path))

    unix_transport, _ = await loop.create_datagram_endpoint(
        lambda: message_handler,
        sock=sock,
    )

    address, password = xmpp_cfg

    xmpp_client = aioxmpp.Client(
        address,
        aioxmpp.make_security_layer(password)
    )

    muc_client = xmpp_client.summon(aioxmpp.MUCClient)

    try:
        async with xmpp_client.connected() as stream:
            rooms = {}
            for muc_info in mucs:
                room, fut = muc_client.join(
                    muc_info["address"],
                    muc_info["nickname"],
                    autorejoin=True,
                )

                await fut
                muc_info["room"] = room
                rooms[muc_info["address"]] = muc_info

            processor = asyncio.ensure_future(process_queue(
                message_queue,
                rooms
            ))

            done, pending = await asyncio.wait(
                [
                    processor,
                    sigint_future,
                ],
                return_when=asyncio.FIRST_COMPLETED,
            )

            if sigint_future in done:
                if not processor.done():
                    processor.cancel()
                try:
                    await processor
                except asyncio.CancelledError:
                    pass
                return

            if processor in done:
                processor.result()
                raise RuntimeError("processor exited early!")
    finally:
        if not sigint_future.done():
            sigint_future.cancel()
        unix_transport.close()


def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-c", "--config",
        type=pathlib.Path,
        default=pathlib.Path.cwd() / "config.toml",
        help="Path to config file (default: ./config.toml)"
    )
    parser.add_argument(
        "-v", "--verbose",
        default=0,
        dest="verbosity",
        action="count",
        help="Increase verbosity (up to -vvv)"
    )

    args = parser.parse_args()

    logging.basicConfig(
        level={
            0: logging.ERROR,
            1: logging.WARNING,
            2: logging.INFO,
        }.get(args.verbosity, logging.DEBUG)
    )

    with args.config.open("r") as f:
        config = toml.load(f)

    address = aioxmpp.JID.fromstr(config["xmpp"]["account"])
    password = config["xmpp"]["password"]

    socket_path = pathlib.Path(config["unix"]["path"])

    mucs = []
    for muc_cfg in config["xmpp"]["muc"]:
        mucs.append(
            {
                "address": aioxmpp.JID.fromstr(muc_cfg["address"]),
                "nickname": muc_cfg.get("nickname", address.localpart),
                "format": muc_cfg.get("format"),
                "required_fields": frozenset(muc_cfg.get("required_fields", [])),
                "head_format": muc_cfg.get("head_format"),
            }
        )

    if socket_path.exists():
        if not socket_path.is_socket():
            raise RuntimeError("{} exists and is not a socket!".format(
                socket_path,
            ))

        # FIXME: do not unlink the socket if it’s still live; abort instead.
        socket_path.unlink()

    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(amain(
            loop,
            (address, password),
            (socket_path, ),
            mucs,
        ))
    finally:
        loop.close()


if __name__ == "__main__":
    main()



########################################################################
# File name: __init__.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.ibb` --- In-Band Bytestreams (:xep:`0047`)
#########################################################

This subpackage provides support for in-band bytestreams.  The
bytestreams are exposed as instances of :class:`asyncio.Transport`,
which allows to speak any protocol implemented as
:class:`asyncio.Protocol` over them.

.. autoclass:: IBBService

.. autoclass:: IBBStanzaType

.. currentmodule:: aioxmpp.ibb.service
.. autoclass:: IBBTransport()

For serializing and deserializing data payloads carried by
:class:`~aioxmpp.Message` stanzas, a descriptor is added to them:

.. attribute:: aioxmpp.Message.xep0047_data
"""
from .xso import IBBStanzaType # NOQA
from .service import IBBService # NOQA

# import aioxmpp.ibb.service



########################################################################
# File name: self_ping.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import random
import time

from datetime import timedelta

import aioxmpp.errors
import aioxmpp.ping
import aioxmpp.stream
import aioxmpp.structs
import aioxmpp.utils


def _apply_jitter(v, amplitude):
    return v * ((random.random() * 2 - 1) * amplitude + 1)


class MUCPinger:
    """
    :param on_fresh: Called when the pinger finds evidence that the user is
        connected
    :param on_exited: Called when the pinger finds evidence that the user is
        disconnected
    :param loop: Event loop to use

    This class manages a coroutine which sends pings to a remote entity and
    interprets the results according to :xep:`410`.

    If the result of a ping indicates that the client is not joined in the MUC
    anymore, `on_exited` is called. If the result of a ping indicates that the
    client is still joined in a MUC, `on_stale` is called. If the result are
    inconclusive, no call is made.

    A ping result does *not* imply a call to :meth:`stop`. The callbacks are
    called on each ping response, thus, on average up to once each
    :attr:`ping_interval` until :meth:`stop` is called.

    Pings are sent once each :attr:`ping_interval` (see there for details on
    the effects of changing the interval while the pinger is running). If
    :attr:`ping_interval` is less than :attr:`ping_timeout`, it is possible
    that multiple pings are in-flight at the same time (this is handled
    correctly). Take into account that resources for tracking up to
    :attr:`ping_timeout` divided by :attr:`ping_interval` IQ responses will be
    required.

    To start the pinger, :meth:`start` must be called.

    .. automethod:: start

    .. automethod:: stop

    .. attribute:: ping_address

        The address pings are sent to.

        This can be changed while the pinger is running. Changes take effect
        when the next ping is sent. Already in-flight pings are not affected.

    .. autoattribute:: ping_interval

    .. autoattribute:: ping_timeout
    """

    def __init__(self, ping_address, client, on_fresh, on_exited, logger, loop):
        super().__init__()
        self.ping_address = ping_address
        self._ping_interval = timedelta(minutes=2)
        self._ping_timeout = timedelta(minutes=8)
        self._client = client
        self._on_fresh = on_fresh
        self._on_exited = on_exited
        self._loop = loop
        self._logger = logger
        self._task = None

    @property
    def ping_interval(self) -> timedelta:
        """
        The interval at which pings are sent.

        While the pinger is running, every `ping_interval` a new ping is
        started. Each ping has its individual :attr:`ping_timeout`.

        Changing this property takes effect after the next ping has been sent.
        Thus, if the :attr:`ping_interval` was set to one day and is then
        changed to one minute, it takes up to a day until the one minute
        interval starts being used.
        """
        return self._ping_interval

    @ping_interval.setter
    def ping_interval(self, value: timedelta):
        # cheap & duck-typey enforcement of timedelta compatibility
        self._ping_interval = value + timedelta()

    @property
    def ping_timeout(self) -> timedelta:
        """
        The maximum time to wait for a reply to a ping.

        Each ping sent by the pinger has its individual timeout, based on this
        property at the time the ping is sent.
        """
        return self._ping_timeout

    @ping_timeout.setter
    def ping_timeout(self, value: timedelta):
        # cheap & duck-typey enforcement of timedelta compatibility
        self._ping_timeout = value + timedelta()

    def start(self):
        """
        Start the pinging coroutine using the client and event loop which was
        passed to the constructor.

        :meth:`start` always behaves as if :meth:`stop` was called right before
        it.
        """
        self._logger.debug("%s: request to start pinger",
                           self.ping_address)
        self.stop()
        self._task = asyncio.ensure_future(self._pinger(), loop=self._loop)

    def stop(self):
        """
        Stop the pinger (if it is running) and discard all data on in-flight
        pings.

        This method will do nothing if the pinger is already stopped. It is
        idempotent.
        """
        self._logger.debug("%s: request to stop pinger",
                           self.ping_address)
        if self._task is None:
            self._logger.debug("%s: already stopped", self.ping_address)
            return

        self._logger.debug("%s: sending cancel signal", self.ping_address)
        self._task.cancel()
        self._task = None

    def _interpret_result(self, task):
        """
        Interpret the result of a ping.

        :param task: The pinger task.

        The result or exception of the `task` is interpreted as follows:

        * :data:`None` result: *positive*
        * :class:`aioxmpp.errors.XMPPError`, ``service-unavailable``:
          *positive*
        * :class:`aioxmpp.errors.XMPPError`, ``feature-not-implemented``:
          *positive*
        * :class:`aioxmpp.errors.XMPPError`, ``item-not-found``: *inconclusive*
        * :class:`aioxmpp.errors.XMPPError`, ``remote-server-not-found``:
          *inconclusive*
        * :class:`aioxmpp.errors.XMPPError`, ``remote-server-timeout``:
          *inconclusive*
        * :class:`aioxmpp.errors.XMPPError`: *negative*
        * :class:`asyncio.TimeoutError`: *inconclusive*
        * Any other exception: *inconclusive*
        """
        if task.exception() is None:
            self._logger.debug("%s: ping reply has no error -> emitting fresh "
                               "event", self.ping_address)
            self._on_fresh()
            return

        exc = task.exception()
        if isinstance(exc, aioxmpp.errors.XMPPError):
            if exc.condition in [
                    aioxmpp.errors.ErrorCondition.SERVICE_UNAVAILABLE,
                    aioxmpp.errors.ErrorCondition.FEATURE_NOT_IMPLEMENTED]:
                self._logger.debug(
                    "%s: ping reply has error indicating freshness: %s",
                    self.ping_address,
                    exc.condition,
                )
                self._on_fresh()
                return

            if exc.condition in [
                    aioxmpp.errors.ErrorCondition.ITEM_NOT_FOUND,
                    aioxmpp.errors.ErrorCondition.REMOTE_SERVER_NOT_FOUND,
                    aioxmpp.errors.ErrorCondition.REMOTE_SERVER_TIMEOUT]:
                self._logger.debug(
                    "%s: ping reply has inconclusive error: %s",
                    self.ping_address,
                    exc.condition,
                )
                return

            self._logger.debug(
                "%s: ping reply has error indicating that the client got "
                "removed: %s",
                self.ping_address,
                exc.condition,
            )
            self._on_exited()

    async def _pinger(self):
        in_flight = []
        next_ping_at = None
        self._logger.debug("%s: pinger booted up", self.ping_address)
        try:
            while True:
                self._logger.debug("%s: pinger loop. interval=%r",
                                   self.ping_address,
                                   self.ping_interval)
                now = time.monotonic()

                ping_interval = self.ping_interval.total_seconds()
                if next_ping_at is None:
                    next_ping_at = now - 1

                timeout = next_ping_at - now

                if timeout <= 0:
                    # do not send pings while the client is in suspended state
                    # (= Stream Management hibernation). This will only add to
                    # the queue for no good reason, we won’t get any reply soon
                    # anyways.
                    if self._client.suspended:
                        self._logger.debug(
                            "%s: omitting self-ping, as the stream is "
                            "currently hibernated",
                            self.ping_address,
                        )
                    else:
                        self._logger.debug(
                            "%s: sending self-ping with timeout %r",
                            self.ping_address,
                            self.ping_timeout,
                        )
                        in_flight.append(asyncio.ensure_future(
                            asyncio.wait_for(
                                aioxmpp.ping.ping(self._client,
                                                  self.ping_address),
                                self.ping_timeout.total_seconds()
                            )
                        ))
                    next_ping_at = now + _apply_jitter(ping_interval, 0.1)
                    timeout = ping_interval

                assert timeout > 0

                if not in_flight:
                    self._logger.debug(
                        "%s: pinger has nothing to do, sleeping for %s",
                        self.ping_address,
                        timeout,
                    )
                    await asyncio.sleep(timeout)
                    continue

                self._logger.debug(
                    "%s: pinger waiting for %d pings for at most %ss",
                    self.ping_address,
                    len(in_flight),
                    timeout,
                )
                done, pending = await asyncio.wait(
                    in_flight,
                    timeout=timeout,
                    return_when=asyncio.FIRST_COMPLETED,
                )

                for fut in done:
                    self._interpret_result(fut)

                in_flight = list(pending)
        finally:
            self._logger.debug("%s: pinger exited", self.ping_address,
                               exc_info=True)
            for fut in in_flight:
                if not fut.done():
                    fut.cancel()


class MUCMonitor:
    """
    :param ping_address: Address to send pings to. Can be changed later with
        :attr:`ping_address`.
    :type ping_address: :class:`aioxmpp.JID`
    :param client: Client to send pings with.
    :type stream: :class:`aioxmpp.stream.StanzaStream`
    :param on_stale: Called when the pinger detects stale state.
    :param on_fresh: Called when the pinger detects fresh state.
    :param on_exited: Called when the pinger detects that the user is not in
        the room anymore.
    :param loop: Event loop to use (defaults to the current event loop)

    .. automethod:: enable

    .. automethod:: disable

    .. automethod:: reset

    .. attribute:: ping_address

        The address to ping.

    .. autoattribute:: stream

    .. autoattribute:: is_stale

    .. autoattribute:: soft_timeout

    .. autoattribute:: hard_timeout

    .. autoattribute:: ping_interval

    .. autoattribute:: ping_timeout
    """

    def __init__(self,
                 ping_address: aioxmpp.structs.JID,
                 client: "aioxmpp.node.Client",
                 on_stale,
                 on_fresh,
                 on_exited,
                 logger,
                 loop=None):
        loop = loop or asyncio.get_event_loop()
        super().__init__()
        self._client = client
        self._is_stale = False
        self.on_stale = on_stale
        self.on_fresh = on_fresh
        self.on_exited = on_exited
        self._soft_timeout = timedelta(minutes=13)
        self._hard_timeout = timedelta(minutes=15)
        self._monitor = aioxmpp.utils.AlivenessMonitor(loop)
        # disable the monitor altogether
        self._monitor.deadtime_hard_limit = None
        self._monitor.deadtime_soft_limit = None
        self._monitor_enabled = False
        self._monitor.on_deadtime_hard_limit_tripped.connect(
            self._hard_limit_tripped
        )
        self._monitor.on_deadtime_soft_limit_tripped.connect(
            self._soft_limit_tripped
        )
        self._logger = logger
        self._pinger = MUCPinger(
            ping_address,
            client,
            self._pinger_fresh_detected,
            self._pinger_exited_detected,
            logger,
            loop,
        )

        self.ping_address = ping_address

    @property
    def is_stale(self) -> bool:
        return self._is_stale

    @property
    def soft_timeout(self) -> timedelta:
        return self._soft_timeout

    @soft_timeout.setter
    def soft_timeout(self, new_value: timedelta):
        # cheap & duck-typey enforcement of timedelta compatibility
        self._soft_timeout = new_value + timedelta()
        if self._monitor_enabled:
            self._monitor.deadtime_soft_limit = new_value

    @property
    def hard_timeout(self) -> timedelta:
        return self._hard_timeout

    @hard_timeout.setter
    def hard_timeout(self, new_value: timedelta):
        # cheap & duck-typey enforcement of timedelta compatibility
        self._hard_timeout = new_value + timedelta()
        if self._monitor_enabled:
            self._monitor.deadtime_hard_limit = new_value

    ping_address = aioxmpp.utils.proxy_property(
        "_pinger",
        "ping_address",
    )

    ping_timeout = aioxmpp.utils.proxy_property(
        "_pinger",
        "ping_timeout",
    )

    ping_interval = aioxmpp.utils.proxy_property(
        "_pinger",
        "ping_interval",
    )

    def enable(self):
        """
        Enable the monitor, if it is not enabled already.

        If the monitor is not already enabled, the aliveness timeouts are reset
        and configured and the stale state is cleared.
        """
        self._logger.debug("%s: request to enable monitoring",
                           self.ping_address)
        if self._monitor_enabled:
            return
        self._is_stale = False
        self._enable_monitor()

    def disable(self):
        """
        Disable the monitor.

        Reset and stop the aliveness timeouts. Cancel and stop pinging.
        """
        self._disable_monitor()
        self._pinger.stop()

    def reset(self):
        """
        Reset the monitor.

        Reset the aliveness timeouts. Clear the stale state. Cancel and stop
        pinging.

        Call `on_fresh` if the stale state was set.
        """
        self._monitor.notify_received()
        self._pinger.stop()
        self._mark_fresh()

    def _mark_stale(self):
        """
        - Emit on_stale if stale flag is cleared
        - Set stale flag
        """
        if not self._is_stale:
            self._logger.debug("%s: transition to stale", self.ping_address)
            self.on_stale()
        self._is_stale = True

    def _mark_fresh(self):
        """
        - Emit on_fresh if stale flag is set
        - Clear stale flag
        """
        if self._is_stale:
            self._logger.debug("%s: transition to fresh", self.ping_address)
            self.on_fresh()
        self._is_stale = False

    def _enable_monitor(self):
        # we need to call notify received *first* to prevent spurious events
        self._monitor.notify_received()
        self._monitor.deadtime_soft_limit = self._soft_timeout
        self._monitor.deadtime_hard_limit = self._hard_timeout
        self._monitor_enabled = True
        self._logger.debug("%s: enabled monitoring: "
                           "soft_timeout=%r "
                           "hard_timeout=%r "
                           "ping_interval=%r "
                           "ping_timeout=%r",
                           self.ping_address,
                           self._soft_timeout,
                           self._hard_timeout,
                           self.ping_interval,
                           self.ping_timeout)

    def _disable_monitor(self):
        # we need to call notify received *first* to prevent spurious events
        self._monitor.notify_received()
        self._monitor.deadtime_soft_limit = None
        self._monitor.deadtime_hard_limit = None
        self._monitor_enabled = False
        self._logger.debug("%s: disabled monitoring", self.ping_address)

    def _pinger_fresh_detected(self):
        self._logger.debug("%s: fresh detected", self.ping_address)
        self._pinger.stop()
        self._monitor.notify_received()
        self._mark_fresh()

    def _pinger_exited_detected(self):
        self._logger.debug("%s: exited detected", self.ping_address)
        self._pinger.stop()
        self.on_exited()

    def _soft_limit_tripped(self):
        self._logger.debug("%s: soft-limit tripped, starting pinger",
                           self.ping_address)
        self._pinger.start()

    def _hard_limit_tripped(self):
        self._logger.debug("%s: hard-limit tripped, marking stale",
                           self.ping_address)
        self._mark_stale()



########################################################################
# File name: test_service.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import unittest
import unittest.mock

from datetime import timedelta

import aioxmpp
import aioxmpp.xso as xso
import aioxmpp.service as service
import aioxmpp.disco.xso as disco_xso
import aioxmpp.ibb as ibb
import aioxmpp.ibb.service as ibb_service
import aioxmpp.ibb.xso as ibb_xso

from aioxmpp.utils import namespaces

from aioxmpp.testutils import (
    make_connected_client,
    CoroutineMock,
    run_coroutine,
)

TEST_FROM = aioxmpp.structs.JID.fromstr("foo@bar.example/baz")
TEST_JID1 = aioxmpp.structs.JID.fromstr("bar@bar.example/baz")
TEST_JID2 = aioxmpp.structs.JID.fromstr("baz@bar.example/baz")


class TestIBBService(unittest.TestCase):

    def setUp(self):
        self.cc = make_connected_client()
        self.cc.local_jid = TEST_FROM
        self.s = ibb.IBBService(
            self.cc,
        )

    def tearDown(self):
        del self.cc
        del self.s

    def test_is_service(self):
        self.assertTrue(issubclass(
            ibb.IBBService,
            aioxmpp.service.Service
        ))

    def test_open_session_block_size_too_large(self):
        protocol = unittest.mock.Mock()
        with self.assertRaisesRegex(ValueError, r"^block_size too large$"):
            run_coroutine(
                self.s.open_session(protocol, TEST_JID1, block_size=(1<<16)))

    def test_open_session_default(self):
        with unittest.mock.patch("aioxmpp.utils.to_nmtoken")\
             as gsi:
            gsi.return_value = "sentinel"
            protocol = unittest.mock.Mock()
            handle, protocl = run_coroutine(
                self.s.open_session(protocol, TEST_JID1)
            )

        self.assertEqual(
            handle.get_extra_info("sid"),
            "sentinel"
        )

        self.assertEqual(
            handle.get_extra_info("block_size"),
            4096
        )

        self.assertEqual(
            handle.get_extra_info("stanza_type"),
            ibb_xso.IBBStanzaType.IQ
        )

        self.assertEqual(
            handle.get_extra_info("peer_jid"),
            TEST_JID1
        )

        (s, (iq,), kwargs), = self.cc.send.mock_calls
        self.assertFalse(kwargs)
        self.assertEqual(iq.to, TEST_JID1)
        self.assertEqual(iq.payload.sid, "sentinel")
        self.assertEqual(iq.payload.stanza, ibb_xso.IBBStanzaType.IQ)
        self.assertEqual(iq.payload.block_size, 4096)


    def test_open_session_non_default(self):
        with unittest.mock.patch("aioxmpp.utils.to_nmtoken")\
             as gsi:
            gsi.return_value = "sentinel"
            protocol = unittest.mock.Mock()
            handle, proto = run_coroutine(
                self.s.open_session(
                    protocol,
                    TEST_JID1,
                    stanza_type=ibb_xso.IBBStanzaType.MESSAGE,
                    block_size=8192,
                )
            )

        self.assertEqual(
            handle.get_extra_info("sid"),
            "sentinel"
        )

        self.assertEqual(
            handle.get_extra_info("block_size"),
            8192
        )

        self.assertEqual(
            handle.get_extra_info("peer_jid"),
            TEST_JID1
        )

        self.assertEqual(
            handle.get_extra_info("stanza_type"),
            ibb_xso.IBBStanzaType.MESSAGE
        )

        (s, (iq,), kwargs), = self.cc.send.mock_calls
        self.assertFalse(kwargs)
        self.assertEqual(iq.to, TEST_JID1)
        self.assertEqual(iq.payload.sid, "sentinel")
        self.assertEqual(iq.payload.stanza, ibb_xso.IBBStanzaType.MESSAGE)
        self.assertEqual(iq.payload.block_size, 8192)

    def test_open_request_enforce_limit(self):
        open_ = ibb_xso.Open()
        open_.sid = "sentinel"
        open_.block_size = 8192
        open_.stanza = ibb_xso.IBBStanzaType.IQ

        iq = aioxmpp.IQ(
            aioxmpp.IQType.SET,
            from_=TEST_JID1,
            payload=open_
        )
        try:
            run_coroutine(self.s._handle_open_request(iq))
        except aioxmpp.errors.XMPPCancelError as e:
            self.assertEqual(
                e.condition,
                aioxmpp.errors.ErrorCondition.NOT_ACCEPTABLE
            )
        else:
            self.fail("missing expected exception")

    def test_open_request_resource_constraint(self):
        open_ = ibb_xso.Open()
        open_.sid = "sentinel"
        open_.block_size = 1<<32
        open_.stanza = ibb_xso.IBBStanzaType.IQ

        iq = aioxmpp.IQ(
            aioxmpp.IQType.SET,
            from_=TEST_JID1,
            payload=open_
        )
        try:
            run_coroutine(self.s._handle_open_request(iq))
        except aioxmpp.errors.XMPPModifyError as e:
            self.assertEqual(
                e.condition,
                aioxmpp.errors.ErrorCondition.RESOURCE_CONSTRAINT
            )
        else:
            self.fail("missing expected exception")

    def test_open_request_allowed_limit(self):
        self.s.session_limit = 1
        self.s.default_protocol_factory = unittest.mock.Mock()

        def on_connection_accepted(transport, protocol):
            pass

        self.s.on_session_accepted.connect(
            on_connection_accepted
        )

        open_ = ibb_xso.Open()
        open_.sid = "sentinel"
        open_.block_size = 8192
        open_.stanza = ibb_xso.IBBStanzaType.IQ

        iq = aioxmpp.IQ(
            aioxmpp.IQType.SET,
            from_=TEST_JID1,
            payload=open_
        )

        run_coroutine(self.s._handle_open_request(iq))

    def test_open_request_expected_session(self):
        protocol_factory = unittest.mock.Mock()
        s_fut = self.s.expect_session(protocol_factory, TEST_JID1, "sentinel")

        open_ = ibb_xso.Open()
        open_.sid = "sentinel"
        open_.block_size = 8192
        open_.stanza = ibb_xso.IBBStanzaType.IQ

        iq = aioxmpp.IQ(
            aioxmpp.IQType.SET,
            from_=TEST_JID1,
            payload=open_
        )
        run_coroutine(self.s._handle_open_request(iq))

        async def await_s_fut():
            return await s_fut

        handle, proto = run_coroutine(await_s_fut())
        self.assertEqual(
            handle.get_extra_info("peer_jid"),
            TEST_JID1
        )
        self.assertEqual(
            handle.get_extra_info("sid"),
            "sentinel"
        )
        self.assertEqual(
            handle.get_extra_info("block_size"),
            8192
        )

        protocol_factory().connection_made.assert_called_with(handle)

    def test_remote_close_unknown_session(self):
        close = ibb_xso.Close()
        close.sid = "quark"
        stanza = aioxmpp.IQ(
            aioxmpp.IQType.SET,
            from_=TEST_JID1,
            to=TEST_FROM,
            payload=close
        )
        try:
            run_coroutine(self.s._handle_close_request(stanza))
        except aioxmpp.errors.XMPPCancelError as e:
            self.assertEqual(
                e.condition,
                aioxmpp.errors.ErrorCondition.ITEM_NOT_FOUND
            )

    def test_handle_open_request_is_iq_handler(self):
        self.assertTrue(
            aioxmpp.service.is_iq_handler(
                aioxmpp.IQType.SET,
                ibb_xso.Open,
                ibb_service.IBBService._handle_open_request,
            )
        )

    def test_handle_close_request_is_iq_handler(self):
        self.assertTrue(
            aioxmpp.service.is_iq_handler(
                aioxmpp.IQType.SET,
                ibb_xso.Close,
                ibb_service.IBBService._handle_close_request,
            )
        )

    def test_handle_data_is_iq_handler(self):
        self.assertTrue(
            aioxmpp.service.is_iq_handler(
                aioxmpp.IQType.SET,
                ibb_xso.Data,
                ibb_service.IBBService._handle_data,
            )
        )

class TestProtocol(asyncio.Protocol):

    def __init__(self):
        self.data = b""
        self.data_received_evt = asyncio.Event()
        self.connection_lost_fut = asyncio.Future()
        self._transport = None

    def connection_made(self, transport):
        self._transport = transport

    def connection_lost(self, e):
        self.connection_lost_fut.set_result(e)

    def pause_writing(self):
        pass

    def resume_writing(self):
        pass

    def data_received(self, data):
        self.data += data
        self.data_received_evt.set()

class TestIBBService_OpenConnection(unittest.TestCase):

    def setUp(self):
        self.cc = make_connected_client()
        self.cc.local_jid = TEST_FROM
        self.s = ibb.IBBService(
            self.cc,
        )
        self.s.initial_wait_time = timedelta(milliseconds=1)
        protocol_factory = unittest.mock.Mock()
        self.handle, self.protocol = run_coroutine(
            self.s.open_session(protocol_factory, TEST_JID1)
        )
        self.cc.send.mock_calls.clear()

    def tearDown(self):
        # ensure clean-up
        self.s._on_stream_destroyed()
        del self.cc
        del self.s

    def test_send(self):
        arguments = []
        async def patched_send(stanza):
            fut.set_result(None)
            arguments.append(stanza)

        for i in range(5):
            fut = asyncio.Future()
            self.cc.send = patched_send
            self.handle.write(b"some data")
            run_coroutine(fut)

        total_content = b""
        for i, iq in enumerate(arguments):
            self.assertEqual(iq.to, TEST_JID1)
            self.assertIsInstance(iq.payload, ibb_xso.Data)
            self.assertEqual(iq.payload.sid, self.handle.get_extra_info("sid"))
            self.assertEqual(iq.payload.seq, i)
            self.assertTrue(iq.payload.content)
            total_content += iq.payload.content
        self.assertEqual(total_content, b"some data"*5)

    def test_receive(self):

        proto = TestProtocol()
        self.handle.set_protocol(proto)
        total_text = b""
        for i in range(5):
            text = "data{}".format(i).encode("us-ascii")
            stanza = aioxmpp.IQ(
                aioxmpp.IQType.SET,
                from_=TEST_JID1,
                to=TEST_FROM,
                payload=ibb_xso.Data(
                    self.handle.get_extra_info("sid"),
                    i,
                    text
                )
            )
            total_text += text
            run_coroutine(self.s._handle_data(stanza))

        self.assertEqual(
            proto.data, total_text
        )

    def test_open_request_established_session(self):
        self.s.session_limit = 2
        self.s.default_protocol_factory = unittest.mock.Mock()
        on_session_accepted = unittest.mock.Mock()
        self.s.on_session_accepted.connect(
            on_session_accepted
        )

        open_ = ibb_xso.Open()
        open_.sid = self.handle.get_extra_info("sid")
        open_.block_size = 4096
        open_.stanza = ibb_xso.IBBStanzaType.IQ

        iq = aioxmpp.IQ(
            aioxmpp.IQType.SET,
            from_=TEST_JID1,
            payload=open_
        )

        try:
            run_coroutine(self.s._handle_open_request(iq))
        except aioxmpp.errors.XMPPCancelError as e:
            self.assertEqual(
                e.condition,
                aioxmpp.errors.ErrorCondition.NOT_ACCEPTABLE
            )

        self.assertEqual(len(on_session_accepted.mock_calls), 0)

    def test_receive_invalid_session(self):
        proto = TestProtocol()
        self.handle.set_protocol(proto)

        stanza = aioxmpp.IQ(
            aioxmpp.IQType.SET,
            from_=TEST_JID1,
            to=TEST_FROM,
            payload=ibb_xso.Data(
                "fnord",
                0,
                b"data",
            )
        )
        try:
            run_coroutine(self.s._handle_data(stanza))
        except aioxmpp.errors.XMPPCancelError as e:
            self.assertEqual(
                e.condition,
                aioxmpp.errors.ErrorCondition.ITEM_NOT_FOUND
            )
        else:
            self.fail("expected exception was not raised")
        self.assertEqual(proto.data, b"")

    def test_receive_invalid_seq(self):
        proto = TestProtocol()
        self.handle.set_protocol(proto)

        stanza = aioxmpp.IQ(
            aioxmpp.IQType.SET,
            from_=TEST_JID1,
            to=TEST_FROM,
            payload=ibb_xso.Data(
                self.handle.get_extra_info("sid"),
                1,
                b"data",
            )
        )
        try:
            run_coroutine(self.s._handle_data(stanza))
        except aioxmpp.errors.XMPPCancelError as e:
            self.assertEqual(
                e.condition,
                aioxmpp.errors.ErrorCondition.UNEXPECTED_REQUEST
            )
        else:
            self.fail("expected exception was not raised")
        self.assertEqual(proto.data, b"")

    def test_wait_error_timeout(self):
        fut = asyncio.Future()
        self.protocol.connection_lost = lambda e: fut.set_result(e)
        self.cc.send.side_effect = aioxmpp.errors.XMPPWaitError(
            aioxmpp.errors.ErrorCondition.RECIPIENT_UNAVAILABLE
        )
        self.handle.write(b"data")
        run_coroutine(fut)
        self.assertEqual(self.handle._retries, self.s.max_retries)
        self.assertFalse(self.s._sessions)
        self.assertIsInstance(fut.result(), asyncio.TimeoutError)

    def test_wait_error_recover(self):
        fut = asyncio.Future()
        fut_close = asyncio.Future()

        def side_effect(*args, **kwargs):
            yield aioxmpp.errors.XMPPWaitError(
                aioxmpp.errors.ErrorCondition.RECIPIENT_UNAVAILABLE
            )
            fut.set_result((args, kwargs))
            yield unittest.mock.DEFAULT
            fut_close.set_result((args, kwargs))
            yield unittest.mock.DEFAULT

        self.cc.send.side_effect = side_effect()
        self.handle.write(b"data")
        run_coroutine(fut)
        self.handle.close()
        run_coroutine(fut_close)
        self.assertEqual(self.handle._wait_time,
                         self.s.initial_wait_time.total_seconds())
        self.assertEqual(self.handle._retries, 0)

    def test_remote_seq_error(self):
        fut = asyncio.Future()
        self.protocol.connection_lost = lambda e: fut.set_result(e)
        self.cc.send.side_effect = aioxmpp.errors.XMPPCancelError(
            aioxmpp.errors.ErrorCondition.UNEXPECTED_REQUEST
        )
        self.handle.write(b"data")
        run_coroutine(fut)
        self.assertIs(fut.result(), self.cc.send.side_effect)
        self.assertFalse(self.s._sessions)

        (_, (iq1,), _), (_, (iq2,), _) = self.cc.send.mock_calls
        self.assertEqual(iq1.type_, aioxmpp.IQType.SET)
        self.assertIsInstance(iq1.payload, ibb_xso.Data)
        self.assertEqual(iq2.type_, aioxmpp.IQType.SET)
        self.assertIsInstance(iq2.payload, ibb_xso.Close)

    def test_remote_abort(self):
        fut = asyncio.Future()
        self.protocol.connection_lost = lambda e: fut.set_result(e)
        self.cc.send.side_effect = aioxmpp.errors.XMPPCancelError(
            aioxmpp.errors.ErrorCondition.ITEM_NOT_FOUND
        )
        self.handle.write(b"data")
        run_coroutine(fut)
        self.assertIs(fut.result(), self.cc.send.side_effect)
        self.assertFalse(self.s._sessions)

        (_, (iq1,), _), (_, (iq2,), _) = self.cc.send.mock_calls
        self.assertEqual(iq1.type_, aioxmpp.IQType.SET)
        self.assertIsInstance(iq1.payload, ibb_xso.Data)
        self.assertEqual(iq2.type_, aioxmpp.IQType.SET)
        self.assertIsInstance(iq2.payload, ibb_xso.Close)


    def test_disconnect(self):
        fut = asyncio.Future()
        self.protocol.connection_lost = lambda e: fut.set_result(e)
        self.cc.send.side_effect = ConnectionError()
        self.handle.write(b"data")
        run_coroutine(fut)
        self.assertIs(fut.result(), self.cc.send.side_effect)
        self.assertFalse(self.s._sessions)

        (_, (iq1,), _), = self.cc.send.mock_calls
        self.assertEqual(iq1.type_, aioxmpp.IQType.SET)
        self.assertIsInstance(iq1.payload, ibb_xso.Data)

    def test_remote_abort_during_close(self):
        fut = asyncio.Future()
        self.protocol.connection_lost = lambda e: fut.set_result(e)
        self.cc.send.side_effect = aioxmpp.errors.XMPPCancelError(
            aioxmpp.errors.ErrorCondition.ITEM_NOT_FOUND
        )
        self.handle.close()
        run_coroutine(fut)
        self.assertIs(fut.result(), self.cc.send.side_effect)

    def test_close(self):
        fut = asyncio.Future()
        self.protocol.connection_lost = lambda e: fut.set_result(None)
        self.handle.close()
        run_coroutine(fut)
        (_, (iq,), _), = self.cc.send.mock_calls
        self.assertEqual(iq.to, TEST_JID1)
        self.assertIsInstance(iq.payload, ibb_xso.Close)
        self.assertEqual(iq.payload.sid, self.handle.get_extra_info("sid"))
        self.assertFalse(self.s._sessions)

    def test_remote_close(self):
        fut = asyncio.Future()
        self.protocol.connection_lost = lambda e: fut.set_result(None)

        close = ibb_xso.Close()
        close.sid = self.handle.get_extra_info("sid")
        stanza = aioxmpp.IQ(
            aioxmpp.IQType.SET,
            from_=TEST_JID1,
            to=TEST_FROM,
            payload=close
        )
        run_coroutine(self.s._handle_close_request(stanza))
        run_coroutine(fut)
        self.assertFalse(self.s._sessions)

    def test_fill_buffer(self):
        pause_fut = asyncio.Future()
        resume_fut = asyncio.Future()
        # hack the block size
        self.handle._block_size = 5
        self.handle.set_write_buffer_limits(20, 10)
        self.protocol.pause_writing = lambda: pause_fut.set_result(True)
        self.protocol.resume_writing = lambda: resume_fut.set_result(True)
        self.handle.write(b" " * 21)
        run_coroutine(pause_fut)
        run_coroutine(resume_fut)
        self.assertLess(len(self.handle._write_buffer), 10)
        self.assertGreater(len(self.handle._write_buffer), 0)

class TestIBBService_OpenConnectionMessage(unittest.TestCase):

    def setUp(self):
        self.cc = make_connected_client()
        self.cc.local_jid = TEST_FROM
        self.s = ibb.IBBService(
            self.cc,
        )
        protocol_factory = unittest.mock.Mock()
        self.handle, self.protocol = run_coroutine(
            self.s.open_session(protocol_factory, TEST_JID1,
                                stanza_type=ibb_xso.IBBStanzaType.MESSAGE)
        )
        self.cc.send.mock_calls.clear()

    def tearDown(self):
        # ensure clean-up
        self.s._on_stream_destroyed()
        del self.cc
        del self.s

    def test_leave_non_protocol_messages_intact(self):
        stanza = aioxmpp.Message(
            aioxmpp.MessageType.NORMAL,
            from_=TEST_JID1,
            to=TEST_FROM,
        )
        self.assertIs(self.s._handle_message(stanza), stanza)

    def test_send(self):

        for i in range(5):
            fut = asyncio.Future()

            async def patched_send(stanza):
                fut.set_result(None)
                patched_send.argument = stanza

            self.cc.send = patched_send
            self.handle.write(b"some data")
            run_coroutine(fut)
            msg = patched_send.argument
            self.assertEqual(msg.to, TEST_JID1)
            self.assertIsInstance(msg.xep0047_data, ibb_xso.Data)
            self.assertEqual(msg.xep0047_data.sid,
                             self.handle.get_extra_info("sid"))
            self.assertEqual(msg.xep0047_data.seq, i)
            self.assertEqual(msg.xep0047_data.content, b"some data")

    # def test_send_long(self):
    #     run_coroutine(self.handle.send(b" " * 4097))
    #     self.assertEqual(
    #         len(self.cc.send.mock_calls),
    #         2
    #     )

    def test_receive(self):

        proto = TestProtocol()
        self.handle.set_protocol(proto)
        total_text = b""
        for i in range(5):
            text = "data{}".format(i).encode("us-ascii")
            stanza = aioxmpp.Message(
                aioxmpp.MessageType.NORMAL,
                from_=TEST_JID1,
                to=TEST_FROM,
            )
            stanza.xep0047_data = ibb_xso.Data(
                self.handle.get_extra_info("sid"),
                i,
                text
            )

            total_text += text
            self.s._handle_message(stanza)

        self.assertEqual(
            proto.data, total_text
        )

    def test_receive_invalid_session(self):
        proto = TestProtocol()
        self.handle.set_protocol(proto)

        stanza = aioxmpp.Message(
            aioxmpp.MessageType.NORMAL,
            from_=TEST_JID1,
            to=TEST_FROM,
        )
        stanza.xep0047_data = ibb_xso.Data(
            "fnord",
            0,
            b"data",
        )

        self.s._handle_message(stanza)
        self.assertEqual(proto.data, b"")

    def test_receive_invalid_seq(self):
        proto = TestProtocol()
        self.handle.set_protocol(proto)

        stanza = aioxmpp.Message(
            aioxmpp.MessageType.NORMAL,
            from_=TEST_JID1,
            to=TEST_FROM,
        )

        stanza.xep0047_data = ibb_xso.Data(
            self.handle.get_extra_info("sid"),
            1,
            b"data",
        )

        self.s._handle_message(stanza)
        self.assertEqual(proto.data, b"")


class TestIBBTransport(unittest.TestCase):

    def setUp(self):
        self.cc = make_connected_client()
        self.cc.local_jid = TEST_FROM
        self.s = ibb.IBBService(
            self.cc,
        )
        protocol_factory = unittest.mock.Mock()
        self.handle, self.protocol = run_coroutine(
            self.s.open_session(protocol_factory, TEST_JID1)
        )
        self.cc.send.mock_calls.clear()

    def tearDown(self):
        # ensure clean-up
        self.s._on_stream_destroyed()
        del self.cc
        del self.s

    def test_set_write_buffer_limits(self):
        self.handle.set_write_buffer_limits(10000, 100)
        self.assertEqual(
            self.handle.get_write_buffer_limits(),
            (100, 10000)
        )

    def test_set_write_buffer_limits_only_high(self):
        low_before, high_before = self.handle.get_write_buffer_limits()
        self.handle.set_write_buffer_limits(high_before*2)
        self.assertEqual(
            self.handle.get_write_buffer_limits(),
            (low_before, high_before*2)
        )

    def test_set_write_buffer_limits_reject_negative(self):
        with self.assertRaises(ValueError):
            self.handle.set_write_buffer_limits(-10, 0)

        with self.assertRaises(ValueError):
            self.handle.set_write_buffer_limits(0, -10)

    def test_set_write_buffer_limits_set_to_zeroe(self):
        self.handle.set_write_buffer_limits(0)
        self.assertEqual(
            self.handle.get_write_buffer_limits(),
            (0, 0)
        )

    def test_set_write_buffer_limits_defaults(self):
        low_before, high_before = self.handle.get_write_buffer_limits()
        self.handle.set_write_buffer_limits(10000, 100)
        self.handle.set_write_buffer_limits()
        self.assertEqual(
            self.handle.get_write_buffer_limits(),
            (low_before, high_before)
        )

    def test_get_write_buffer_size(self):
        self.handle.write(b" "*10)
        self.assertEqual(self.handle.get_write_buffer_size(), 10)

    def test_get_protocol(self):
        self.assertIs(self.handle.get_protocol(), self.protocol)

    def test_pause_and_resume_reading(self):
        self.handle.pause_reading()
        self.handle._data_received(b"foo")
        self.handle._data_received(b"bar")
        self.assertSequenceEqual(
            self.protocol.data_received.mock_calls,
            []
        )
        self.handle.resume_reading()
        self.assertSequenceEqual(
            self.protocol.data_received.mock_calls,
            [unittest.mock.call(b"foobar")]
        )
